qx.Class.define("pf.Plasma",
{
    extend: qx.core.Object,
    
    members:
    {
        R0: 0.5, // 0.3/0.7/1 ?
        
        e: null,
        i: null,
        r: null,
        P_r: 0,
        B  : 0,
        
        getTransitionEnergy: function()
        {
            var n_u = this.n_u;
            var n_l = this.n_l;
            var Z_core = this.r.getQ() + 1;
            var m_core = this.r.getM();
            
            return pf.base.Bohr.transitionEnergy(Z_core, m_core, n_u, n_l);
        },
        
        getFineStructure: function()
        {
            var n_u = this.n_u;
            var n_l = this.n_l;
            var Z_core = this.r.getQ() + 1;
            
            return pf.base.Bohr.transitionFineStructure(Z_core, n_u, n_l);
        },
        
        getMinEnergyDistance: function()
        {
            var n;
            if (this.n_u < this.n_l) {
                n = this.n_l;
            } else {
                n = this.n_u;
            }
            
            var Z_core = this.r.getQ() + 1;
            var m_core = this.r.getM();
            
            return pf.base.Bohr.transitionEnergy(Z_core, m_core, n + 1, n);
        },
        
        getNaturalWidth: function()
        {
            var n_u = this.n_u;
            var n_l = this.n_l;
            var Z_core = this.r.getQ() + 1;
            
            return Math.pow(Z_core, 4)*
              (pf.base.Bohr.LymanA(n_u) + pf.base.Bohr.LymanA(n_l));
        },
        
        getDopplerHwhm: function()
        {
            var v   = this.r.getThermalVelocity();
            var nu  = this.getTransitionEnergy();
            
            return Math.sqrt(2*Math.log(2))*nu*v/pf.base.Bohr.c;
        },
        
        getMicrofieldFrequency: function(s)
        {
            var r = s.getTypicalDistance();
            var v_r = this.r.getThermalVelocity();
            var v_s = s.getThermalVelocity();
            var v = Math.sqrt(v_r*v_r + v_s*v_s);
            
            return v/r;
        },
        
        getFullDebyeLength: function(s)
        {
            var tmp = 0;
            var m = s.getM();
            
            var sa = new Array(this.e, this.i, this.r);
            for (var i = 0; i < 3; i++) {
                var os = sa[i];
                if (os.getM() <= m) {
                    var ld = os.getDebyeLength();
                    if (ld) {
                        tmp += 1/(ld*ld);
                    }
                }
                
            }
            
            return 1/Math.sqrt(tmp);
        },
        
        getPpiFactor: function(s)
        {
            var dl = this.getFullDebyeLength(s);
            var r1 = s.getTypicalDistance();
            var r;
            if (isFinite(dl) && isFinite(r1)) {
                r = r1/dl;
            } else {
                r = 0;
            }
            
            return (1 + r)*Math.exp(-r);
            // return 1 - 0.6*r
        },
        
        getRpiFactor: function(s)
        {
            var r1 = s.getTypicalDistance();
            var q_r = this.r.getQ();
            var q_s = s.getQ();
            if (q_r == 0 || q_s == 0) {
                return 1;
            }
             
            var r_m = s.getMinApproachDistance()*q_r/q_s;
            return Math.exp(-r_m/r1);
        },
        
        getDelta0: function(s)
        {
            var n_u = this.n_u;
            var n_l = this.n_l;
            var Z_core = this.r.getQ() + 1;
            var ef = s.getHoltsmarkField();
            
            return 3/2*(n_u*n_u - n_l*n_l)/Z_core*ef;
        },
        
        getQsHwhm: function(s)
        {
            var n_u = this.n_u;
            var n_l = this.n_l;

            var ppi = this.getPpiFactor(s);
            var rpi = this.getRpiFactor(s);
            
            // 1.4385 is S_1 HWHM
            var hwhm = 1.4385*this.getDelta0(s)*ppi*rpi;
            
            // rude correction for alpha lines
            if (n_u - n_l == 1) {
                hwhm /= 2;
            }
            
            return hwhm;
        },
        
        getDynHwhm: function(s)
        {
            // neglect dynamic PP effects
            var ppi = 1.0;
            // assume dynamic RP effect == static one
            var rpi = this.getRpiFactor(s);

            var fwhm = this.getMicrofieldFrequency(s)*ppi*rpi;
            
            return fwhm/2;
        },
        
        getStarkRatio: function(s)
        {
            var R;
            var w_qs = this.getQsHwhm(s);
            if (w_qs == 0) {
                R = 0;
            } else {
                var w_dyn = this.getDynHwhm(s);
                R = w_qs/w_dyn;
            }
            
            return R;
        },
        
        getStarkQuasistaticity: function(s)
        {
            var R = this.getStarkRatio(s);
            return R/(R + this.R0);
        },
        
        getStarkHwhm: function()
        {
            var w = 0;
            
            var sa = new Array(this.e, this.i, this.r);
            for (var i = 0; i < 3; i++) {
                var s = sa[i];
                var qs = this.getQsHwhm(s);
                var f  = this.getStarkQuasistaticity(s);

                w += Math.pow(f*qs, 3/2);
            }
            
            return Math.pow(w, 2/3);
        },
        
        getZeemanSplitting: function()
        {
            return pf.base.Bohr.mu_B*this.B;
        },

        getTotalFwhm: function()
        {
            var s_hwhm = this.getStarkHwhm();
            var z_hwhm = this.getZeemanSplitting();
            var d_hwhm = this.getDopplerHwhm();
            return 2*Math.sqrt(s_hwhm*s_hwhm + z_hwhm*z_hwhm + d_hwhm*d_hwhm);
        },

        getBremsstrahlungLosses: function()
        {
            var z_i = this.i.getQ();
            var z_r = this.r.getQ();
            var v_e = this.e.getThermalVelocity();
            
            return 8*Math.PI/(3*Math.pow(pf.base.Bohr.c, 3))*v_e*this.e.getN()*
                (this.i.getN()*z_i*z_i + this.r.getN()*z_r*z_r);
        },
        
        getBremsstrahlungSpectralDensity: function()
        {
            var photon_en = this.getTransitionEnergy();
            return this.getBremsstrahlungLosses()/this.e.getT()*
                Math.exp(-photon_en/this.e.getT());
        },
        
        getFreeBoundLosses: function()
        {
            var z_i = this.i.getQ();
            var z_r = this.r.getQ();
            var v_e = this.e.getThermalVelocity();
            var m_r = this.r.getM();

            return -8*Math.PI/(3*Math.pow(pf.base.Bohr.c, 3))*v_e*this.e.getN()*
                (this.i.getN()*z_i*z_i*pf.base.Bohr.bindingEnergy(z_i, m_r, 1) +
                 this.r.getN()*z_r*z_r*pf.base.Bohr.bindingEnergy(z_r, m_r, 1))/
                 this.e.getT();
        },
        
        getFreeBoundSpectralDensity: function()
        {
            var photon_en = this.getTransitionEnergy();
            return this.getFreeBoundLosses()/this.e.getT()*
                Math.exp(-photon_en/this.e.getT());
        },
        
        getBoundBoundLosses: function()
        {
            var Z_core = this.r.getQ() + 1;
            var m_core = this.r.getM();
            var photon_en =
                pf.base.Bohr.transitionEnergy(Z_core, m_core, 2, 1);
            
            // FIXME!!
            var level_pop = Math.exp(-photon_en/this.e.getT());
            
            return level_pop*Math.pow(Z_core, 4)*pf.base.Bohr.LymanA(2)*
                photon_en*this.r.getN();
        },
        
        getPressure: function()
        {
            return this.e.getPressure() +
                   this.i.getPressure() +
                   this.r.getPressure();
        },
        
        getMagneticFieldPressure: function()
        {
            return this.B*this.B/(8*Math.PI);
        },
        
        getBeta: function()
        {
            return this.getPressure()/this.getMagneticFieldPressure();
        },
        
        getFermiEnergy: function()
        {
            return pf.base.Thermo.fermiEnergy(this.e.getN());
        },
        
        getAlfvenVelocity: function()
        {
            var v_m1;
            
            v_m1 = 1/this.e.getAlfvenVelocity(this.B) +
                   1/this.i.getAlfvenVelocity(this.B);
        
            if (this.r.getQ() != 0) {
                v_m1 += 1/this.r.getAlfvenVelocity(this.B);
            }
            
            return 1/v_m1;
        },

        getCoulombLog: function(s1, s2)
        {
            var lambda;
            
            var m1 = s1.getM();
            var m2 = s2.getM();
            var v1 = s1.getThermalVelocity();
            var v2 = s2.getThermalVelocity();
            var q1 = s1.getQ();
            var q2 = s2.getQ();
            
            // kinetic energy in the CoM frame
            var Kp = 3*m1*m2/(2*(m1 + m2))*(v1 + v2)*(v1 + v2);
            
            var b_min = Math.abs(q1*q1/Kp);
            var b_max = this.getFullDebyeLength(s1) +
                        this.getFullDebyeLength(s2);
            
            lambda = Math.log(b_max/b_min);
            if (lambda < 1.0) {
                lambda = 1.0;
            }
            
            return lambda;
        },

        getRelaxationRate: function(s1, s2)
        {
            var m1 = s1.getM();
            var m2 = s2.getM();
            var v1 = s1.getThermalVelocity();
            var v2 = s2.getThermalVelocity();
            var q1 = s1.getQ();
            var q2 = s2.getQ();
            
            var n2 = s2.getN();
            
            var lambda = this.getCoulombLog(s1, s2);
            
            // LL X (42,5) with the missing factor of 3
            return Math.sqrt(128*Math.PI)/3*q1*q1*q2*q2*n2*lambda/
                   (Math.pow(v1*v1 + v2*v2, 3/2)*m1*m2);
        },

        getDreicerField: function()
        {
            var n_e = this.e.getN();
            var T_e = this.e.getT();
            var lambda = this.getCoulombLog(this.e, this.e);
            
            return 4*Math.PI*lambda*n_e/T_e;
        },
        
        getClassicalIonizationField: function()
        {
            var Z_core = this.r.getQ() + 1;
            
            return pf.base.Bohr.classicalIonizationField(Z_core, this.n_u);
        },
        
        getInglisTellerField: function()
        {
            var Z_core = this.r.getQ() + 1;
            var n = this.n_u;
            
            return Math.pow(Z_core, 3)*
                (2*n + 1)/(6*Math.pow(n, 4)*(n + 1)*(n + 1));
        },
        

        checkParameters: function()
        {
            if ((this.P_r < 0.0 || this.P_r > 1.0) ||
                this.i.getQ() < 1 ||
                this.P_r == 1.0 && this.r.getQ() == 0) {
                return false;
            } else {
                return true;
            }
        },

        updateParameters: function()
        {
            // Safety measures
            if (this.checkParameters() != true) {
                return false;
            }
            
            var N_e = this.e.getN();
            var N_i, N_r;
            
            if (this.P_r < 1.0) {
                N_i = N_e*(1.0 - this.P_r)/
                    (this.i.getQ()*(1.0 - this.P_r) + this.r.getQ()*this.P_r);
                
                N_r = N_i*this.P_r/(1.0 - this.P_r);
            } else {
                N_i = 0.0;
                N_r = N_e/this.r.getQ();
            }
            
            this.i.setN(N_i);
            this.r.setN(N_r);
            
            return true;
        }
    },
    
    construct: function()
    {
        var m_p = pf.base.Bohr.m_p;
        var V0  = pf.base.Bohr.V0;
        var E0  = pf.base.Bohr.E0;
        
        // Defaults: 1e16 1/cc, 1 eV
        this.r = new pf.base.Species("r", m_p,  0,       0, 1.0/E0, null);
        this.e = new pf.base.Species("e", 1.0, -1, 1e16*V0, 1.0/E0, this.r);
        this.i = new pf.base.Species("i", m_p, +1, 1e16*V0, 1.0/E0, this.r);

        this.P_r = 0;

        this.B   = 0;

        this.n_u = 2;
        this.n_l = 1;
    }
});
